
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/event_groups.h"
#include "freertos/queue.h"
#include "nvs_flash.h"
#include "esp_log.h"
#include "driver/gpio.h"


#include "appble_central.h"
#define ON_BOARD_LED			GPIO_NUM_18


const char * APPBLE_TAG = "AppBleCentral";

uint8_t g_abcreq[]   = {'A',0x2C,'B',0x2C,'C',0x2C,'D',0x2C};
uint8_t g_123req[] = {'1',0x2C,'2',0x2C,'3',0x2C,'4',0x2C};
uint8_t g_configreq[] = {'A',0x2C,'B',0x2C,'C',0x2C,'D',0x2C,'1',0x2C,'2',0x2C,'3',0x2C,'4',0x2C};

QueueHandle_t xAppBleEventQueue;
TimerHandle_t xAppBleTimer;
static uint16_t g_readReqCntr, g_writeReqCntr;

void AppBle_Reset_Timer(void);




bool SendEventToAppBleTask(AppBleEventStruct * xEvents){
	BaseType_t xQueueSendStatus;

	xQueueSendStatus = xQueueSendToBack(xAppBleEventQueue, xEvents, pdMS_TO_TICKS(100));
	if(xQueueSendStatus != pdPASS){
		return false;
	}
	return true;
}

void AppBle_Reset_Timer(void){

    if (xTimerReset(xAppBleTimer, pdMS_TO_TICKS(1000)) == pdPASS) {
    	ESP_LOGI(APPBLE_TAG, "AppBle Timer Reset!");
    }
}

void AppBleTimerCallback(TimerHandle_t timerhandle){

	AppBleDataStruct * data = 	(AppBleDataStruct *)pvTimerGetTimerID(timerhandle);
	if(data->isConnectedFlag){
		if(data->ledblink_cntr){
			data->ledblink_cntr = 0;
		} else {
			data->ledblink_cntr = 1;
		}
		gpio_set_level(ON_BOARD_LED, data->ledblink_cntr);
		//read write request triggers here.
		if(g_readReqCntr++ >= APPBLE_READ_REQ_TIMEOUT){
			g_readReqCntr = 0;
			AppBleEventStruct bleEvent;
			bleEvent.events = eAppBleReadEvent;
			bleEvent.pvData = NULL;
			if(false == SendEventToAppBleTask(&bleEvent)){
				ESP_LOGE(APPBLE_TAG, "Failed sending read event from timer!");
			}
		}

		if(g_writeReqCntr++ >= APPBLE_WRITE_REQ_TIMEOUT){
			g_writeReqCntr = 0;
			AppBleEventStruct bleEvent;
			bleEvent.events = eAppBleWriteEvent;
			bleEvent.pvData = NULL;
			if(false == SendEventToAppBleTask(&bleEvent)){
				ESP_LOGE(APPBLE_TAG, "Failed sending write event from timer!");
			}
		}

	}

}

void ble_manager(void * pvParamter)
{
	AppBleEventStruct appBleEventData;
	AppBleDataStruct AppBleData;
	BaseType_t xStatus;
	uint8_t write_state = 0, read_state = 0;

	gpio_set_direction(ON_BOARD_LED, GPIO_MODE_OUTPUT);
	gpio_set_level(ON_BOARD_LED, 1);
	ble_app_main();
    memset(&appBleEventData, 0, sizeof(appBleEventData));
    memset(&AppBleData, 0, sizeof(AppBleData));
    xAppBleTimer = xTimerCreate("App Timer", pdMS_TO_TICKS(APP_BLE_TIMER_INTERVAL), pdTRUE, &AppBleData, AppBleTimerCallback);
    xTimerStart(xAppBleTimer, pdMS_TO_TICKS(100));
	while(1){
//		ESP_LOGI(APPBLE_TAG, "App Waiting for Events");

		xStatus = xQueueReceive(xAppBleEventQueue, &appBleEventData, portMAX_DELAY);

		if(xStatus != pdPASS){
			//Empty Queue Timeout Reached!!
			ESP_LOGI(APPBLE_TAG, "App Waiting for Events timeout reached");
			continue;
		}

		switch(appBleEventData.events){

			case  eAppBleStartEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleStartEvent");
				ble_app_main();
			}
			break;

			case  eAppBleConnectEvent:
			{
				AppBleData.isConnectedFlag = true;
				AppBleData.conn_handle = (uint16_t *)appBleEventData.pvData;
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleConnectEvent :%d", AppBleData.conn_handle );
				//reset timer here for led control
			}
			break;

			case  eAppBleDisconnectEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleDisconnectEvent");
				AppBleData.isConnectedFlag = false;
				AppBleData.isComSecuredFlag = false;
				//reset timer here for led control
			}
			break;

			case eAppBleCommSecuredEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleDisconnectEvent");
				AppBleData.isComSecuredFlag = true;
			}
			break;

			case  eAppBleReadEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleReadEvent-%d", read_state);
				if((!AppBleData.isConnectedFlag) || (!AppBleData.isComSecuredFlag))
					break;
				switch(read_state){
					case 0:
					{
						read_ble(AppBleData.conn_handle, BLE_BATTERY_CHR);
						read_state++;
					}
					break;

					case 1:
					{
						read_ble(AppBleData.conn_handle, BLE_ACTIVITY_CHR);
						read_state++;
					}
					break;

					case 2:
					{
						read_ble(AppBleData.conn_handle, BLE_CONFIG_CHR);
						read_state = 0;
					}
					break;

					default:
					{
						read_state=0;
					}
					break;
				}

			}
			break;

			case  eAppBleWriteEvent:
			{
				int st = 0;
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleWriteEvent-%d", write_state);

				if((!AppBleData.isConnectedFlag) || (!AppBleData.isComSecuredFlag))
					break;

				switch(write_state){
					case 0:
					{
						st = write_ble(AppBleData.conn_handle, BLE_CONFIG_CHR, g_configreq, sizeof(g_configreq));
						write_state++;
					}
					break;

					case 1:
					{
						st = write_ble(AppBleData.conn_handle, BLE_ACTIVITY_CHR, g_123req, sizeof(g_123req));
						write_state++;
					}
					break;

					case 2:
					{
						st = write_ble(AppBleData.conn_handle, BLE_BATTERY_CHR, g_abcreq, sizeof(g_abcreq));
						write_state=0;
					}
					break;

					default:
					{
						st = 1;
						write_state=0;
					}
					break;
				}

				if(st){
					//error
					ESP_LOGE(APPBLE_TAG, "Unable to proceed for write request!");
				}
			}
			break;

			case  eAppBleNotificationRxEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleNotificationRxEvent");
				AppBleRxPacket * rxpacket = (AppBleRxPacket * )appBleEventData.pvData;
				if(rxpacket->status){
					//error
				} else {
					// data received
					ESP_LOGI(APPBLE_TAG, "Received Data(%d)", rxpacket->datalen);
				}
				free(rxpacket);
			}
			break;

			case  eAppBleWriteCompleteEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleWriteCompleteEvent");
			}
			break;

			case  eAppBleReadCompleteEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleReadCompleteEvent");
				AppBleRxPacket * rxpacket = (AppBleRxPacket * )appBleEventData.pvData;
				if(rxpacket->status){
					//error
					ESP_LOGI(APPBLE_TAG, "Received Data failed");
				} else {
					// data received
					ESP_LOGI(APPBLE_TAG, "Received Data(%d),packet:%X", rxpacket->datalen, (unsigned int)rxpacket);
				}
				free(rxpacket);

			}
			break;


			case  eAppBleStopEvent:
			{
				ESP_LOGI(APPBLE_TAG, "Event Detected: eAppBleStopEvent");
				blecent_stop();
			}
			break;

			default:
			{

			}
			break;
		}
	}



}

void app_main(void){

	esp_err_t ret = nvs_flash_init();
	if  (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND) {
		ESP_ERROR_CHECK(nvs_flash_erase());
		ret = nvs_flash_init();
	}
	ESP_ERROR_CHECK(ret);

	xAppBleEventQueue = xQueueCreate(3, sizeof(AppBleEventStruct));


	xTaskCreate(&ble_manager, "ble manager", 4096, NULL, 5, NULL);

}
